# Code dedicated to an cache around obsolescence property
#
# This module content aims at being upstreamed.
#
# Copyright 2017 Pierre-Yves David <pierre-yves.david@ens-lyon.org>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from . import (
    exthelper,
)

eh = exthelper.exthelper()

try:
    obsstorefilecache = localrepo.localrepository.obsstore
except AttributeError:
    # XXX hg-3.8 compat
    #
    # mercurial 3.8 has issue with accessing file cache property from their
    # cache. This is fix by 36fbd72c2f39fef8ad52d7c559906c2bc388760c in core
    # and shipped in 3.9
    obsstorefilecache = localrepo.localrepository.__dict__['obsstore']

# obsstore is a filecache so we have do to some spacial dancing
@eh.wrapfunction(obsstorefilecache, 'func')
def obsstorewithcache(orig, repo):
    obsstore = orig(repo)

    class cachekeyobsstore(obsstore.__class__):

        _obskeysize = 200

        def cachekey(self, index=None):
            """return (current-length, cachekey)

            'current-length': is the current length of the obsstore storage file,
            'cachekey' is the hash of the last 200 bytes ending at 'index'.

            if 'index' is unspecified, current obsstore length is used.
            Cacheckey will be set to null id if the obstore is empty.

            If the index specified is higher than the current obsstore file
            length, cachekey will be set to None."""
            try:
                with self.svfs('obsstore') as obsfile:
                    obsfile.seek(0, 2)
                    obsstoresize = obsfile.tell()
                    if index is None:
                        index = obsstoresize
                    elif obsstoresize < index:
                        return obsstoresize, None
                    actualsize = min(index, self._obskeysize)
                    obsfile.seek(index - actualsize, 0)
                    keydata = obsfile.read(actualsize)
                return obsstoresize, hashlib.sha1(keydata).digest()
            except (OSError, IOError) as e:
                if e.errno != errno.ENOENT:
                    raise
                return 0, node.nullid

    obsstore.__class__ = cachekeyobsstore

    return obsstore

def getcachekey(repo):
    """get a cache key covering the changesets and obsmarkers content

    IT contains the following data. Combined with 'upgradeneeded' it allows to
    do iterative upgrade for cache depending of theses two data.

    The cache key parts are"
    - tip-rev,
    - tip-node,
    - obsstore-length (nb markers),
    - obsstore-file-size (in bytes),
    - obsstore "cache key"
    """
    assert repo.filtername is None
    cl = repo.changelog
    index, key = repo.obsstore.cachekey()
    tiprev = len(cl) - 1
    return (tiprev,
            cl.node(tiprev),
            len(repo.obsstore),
            index,
            key)

def upgradeneeded(repo, key):
    """return (valid, start-rev, start-obs-idx)

    'valid': is "False" if older cache value needs invalidation,

    'start-rev': first revision not in the cache. None if cache is up to date,

    'start-obs-idx': index of the first obs-markers not in the cache. None is
                     up to date.
    """

    # XXX ideally, this function would return a bounded amount of changeset and
    # obsmarkers and the associated new cache key. Otherwise we are exposed to
    # a race condition between the time the cache is updated and the new cache
    # key is computed. (however, we do not want to compute the full new cache
    # key in all case because we want to skip reading the obsstore content. We
    # could have a smarter implementation here.
    #
    # In pratice the cache should be updated after each transaction within a
    # lock. So we should be fine. We could enforce this with a new repository
    # requirement (or fix the race, that is not too hard).
    invalid = (False, 0, 0)
    if key is None:
        return invalid

    ### Is the cache valid ?
    keytiprev, keytipnode, keyobslength, keyobssize, keyobskey = key
    # check for changelog strip
    cl = repo.changelog
    tiprev = len(cl) - 1
    if (tiprev < keytiprev
            or cl.node(keytiprev) != keytipnode):
        return invalid
    # check for obsstore strip
    obssize, obskey = repo.obsstore.cachekey(index=keyobssize)
    if obskey != keyobskey:
        return invalid

    ### cache is valid, is there anything to update

    # any new changesets ?
    startrev = None
    if keytiprev < tiprev:
        startrev = keytiprev + 1

    # any new markers
    startidx = None
    if keyobssize < obssize:
        startidx = keyobslength

    return True, startrev, startidx
